#ifndef BUFFERMANAGER_H 
#define BUFFERMANAGER_H

#include <new>

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <sys/mman.h>

#include <errno.h>

#include "ringbuffer.h"

template <class T>
class BufferManager : public RingBuffer <T*>
{
    private:
        T *buffers;
        size_t bufnum;

    public:
        BufferManager(uint32_t bufnum) : RingBuffer<T*>(bufnum)
        {
            uint32_t i;

            this->bufnum = bufnum;

            this->buffers = (T*)mmap((void *) 0, bufnum * sizeof(T), 
                            PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANONYMOUS, (int) -1, (off_t) 0);

            for (i = 0; i < bufnum; i++)
                new (&(this->buffers[i])) T();

//            ERROR_DEBUG(this->buffers == MAP_FAILED, HALT, "Mapping failed in BufferManager!\n");

            if (this->buffers == MAP_FAILED)
            {
               perror("BufferManager"); 
                fprintf(stderr, "Mapping failed in BufferManager: %u\n", errno);
                exit(-1);
            }

            for (i = 0; i < bufnum; i++)
                put(this->buffers + i);
        }

        ~BufferManager()
        {
            munmap(buffers,bufnum * sizeof(T));
        }

};

#endif

