#ifndef SINKSOURCE_H
#define SINKSOURCE_H

#include <stdlib.h>
#include <stdint.h>

template <typename T>
class Sink
{
    public:
        virtual void put(T t) = 0;

        virtual ~Sink(){}
};

template <typename T>
class Source
{
    public:
        virtual T get() = 0;
        virtual bool get(T *buf, uint32_t secs, uint32_t nsecs) = 0;
        virtual void unget(T data) = 0;

        virtual ~Source(){}
};

template <typename T>
class Pipe : public Source<T>, public Sink<T> { };

template <typename T>
class Endpoint
{
    public:
        Source<T> *source;
        Sink<T> *sink;

        Endpoint(Source<T> *source, Sink<T> *sink)
        {
            this->source = source;
            this->sink = sink;
        }
};

template <typename T>
class SourceList
{
    public:
        virtual Source<T> *getSource(int index) = 0;
};

template <typename T>
class SinkList
{
    public:
        virtual Sink<T> *getSink(int index) = 0;
};

template <typename T>
class PipeList : public SourceList<T>, public SinkList<T>
{
    private:
        Pipe<T> **pipe;

        int max_num;
        int num;

    public:
        inline Source<T> *getSource(int index)
        {
            return pipe[index];
        };

        inline Sink<T> *getSink(int index)
        {
            return pipe[index];
        };

        ~PipeList()
        {
            if (pipe != NULL)
                free(pipe);
        }

        PipeList(int num)
        {
            this->max_num = num;
            this->num = 0;

            pipe = (Pipe<T> **)malloc(num * sizeof(Pipe<T> *));
        }

        void add(Pipe<T> *item)
        {
            if (num < max_num)
            {
                pipe[num] = item;
                num++;
            }
            else throw -1;
        }

        template <typename U>
            void deleteContent()
            {
                for (int i = 0; i < num; i++)
                    delete (U*)pipe[i];
            }
};

#endif

