#include <stdio.h>

#include "thread.h"
#include "task.h"

Thread::Thread(Task *task)
{
    this->thread = -1;
    this->task = task;
}

void *thread_func(void* arg)
{
    Thread *thread = (Thread*) arg;

    thread->task->run(thread->argv);

    return NULL;
}

void Thread::start(void *argv)
{
    this->argv = argv;
    pthread_create(&(this->thread), NULL, thread_func, this);
}

void Thread::join()
{
    if (pthread_join(this->thread, NULL) != 0)
        perror("pthread_join: ");

    pthread_detach(this->thread);
}

/************************ ThreadGroup **********************/

ThreadGroup::ThreadGroup(Tasks *task)
{
    this->num_threads = task->getNumTasks();

    this->thread = (Thread**)malloc(this->num_threads * sizeof(Thread*));

    for (int i = 0; i < num_threads; i++)
        this->thread[i] = new Thread(task->getTask(i));
}

ThreadGroup::~ThreadGroup()
{
    for (int i = 0; i < num_threads; i++)
        delete this->thread[i];

    free(this->thread);
}

void ThreadGroup::start(void *argv)
{
    for (int i = 0; i < this->num_threads; i++)
        this->thread[i]->start(argv);
}

void ThreadGroup::join()
{
    for (int i = 0; i < this->num_threads; i++)
        this->thread[i]->join();
}


