#ifndef ENTRY_H
#define ENTRY_H

#include <inttypes.h>
#include <string.h>
#include "params.h"
#include <stdio.h>
#include "include/debug.h"
//#include "uint128.h"

//#include "matrix-tweaked.h"

extern uint8_t bytematrix[MATRIX_BYTES];

/* 
 * 0    <= pos1 < 2^14 
 * pos1 <  pos2 < 2^14
 * 0    <= pos3 < 2^13 pos3 always describes the position from the halve-block
 */
template<typename T> inline T get_part_value(const uint16_t pos1, const uint16_t pos2, const uint16_t pos3, 
        const unsigned int listid, const unsigned int byte_offset)
{
    if(listid % 2)
    {
        /* Odd list id */
        return (*(T*)(bytematrix + ((listid/2)*49152 + 24576 + pos3)*COLUMN_BYTES + byte_offset)^ 
                *(T*)(bytematrix + ((listid/2)*49152 + 32768 + pos1)*COLUMN_BYTES + byte_offset)^ 
                *(T*)(bytematrix + ((listid/2)*49152 + 32768 + pos2)*COLUMN_BYTES + byte_offset));
    }
    else
    {
        /* Even list id */
        return (*(T*)(bytematrix + ((listid/2)*49152 + pos1)*COLUMN_BYTES + byte_offset) ^ 
                *(T*)(bytematrix + ((listid/2)*49152 + pos2)*COLUMN_BYTES + byte_offset) ^ 
                *(T*)(bytematrix + ((listid/2)*49152 + 16384 + pos3)*COLUMN_BYTES + byte_offset));
    }
}

static const int MultiplyDeBruijnBitPosition[32] = 
{
    0, 1, 28, 2, 29, 14, 24, 3, 30, 22, 20, 15, 25, 17, 4, 8, 
    31, 27, 13, 23, 21, 19, 16, 7, 26, 12, 18, 6, 11, 5, 10, 9
};

#define ld(v) (MultiplyDeBruijnBitPosition[(uint32_t)((v) * 0x077CB531U) >> 27])

#define MIN(a,b) (((a)<(b))?(a):(b))
#define NODEBITS(nlists) MIN((ld(NNODES/nlists)+1),ld(NNODES))

#define LEVEL_STARTBIT(level) (level * SORT_LEVEL_RANGE + CLIPPING)
#define PRESORT_STARTBIT(level,nlists) (LEVEL_STARTBIT(level) + NODEBITS(nlists))
#define MERGE_STARTBIT(level,nlists) (PRESORT_STARTBIT(level,nlists) + PRESORTED)
#define CMP_NBITS(level,nlists) (SORT_LEVEL_RANGE - PRESORTED - NODEBITS(nlists))

#define CMP_MASK(level,nlists)((1 << CMP_NBITS(level,nlists)) - 1)

/*
   ((level == MAX_LEVEL) ? \
   ((1 << (ALL_BITS - MERGE_STARTBIT(level,nlists))) - 1) : \
   ((1 << CMP_NBITS(level,nlists)) - 1))
   */

#define ALL_BYTES (ALL_BITS >> 3)

#define MAX_LISTS(level) (1 << (MAX_LEVEL - level))

// CHECK: Is the handling using MAX_LISTS ok?
#define LEVEL_2_BITS (ALL_BITS - LEVEL_STARTBIT(2) - PRESORTED - 2)
//(ALL_BITS - MERGE_STARTBIT(2,MAX_LISTS(2)))
#define LEVEL_2_BYTES ((LEVEL_2_BITS >> 3) + ((LEVEL_2_BITS & 0x7) > 0 ? 1 : 0))

#define LEVEL_3_BITS (ALL_BITS - LEVEL_STARTBIT(3) - PRESORTED - 2)
//(ALL_BITS - MERGE_STARTBIT(3,MAX_LISTS(3)))
#define LEVEL_3_BYTES ((LEVEL_3_BITS >> 3) + ((LEVEL_3_BITS & 0x7) > 0 ? 1 : 0))

template<unsigned level>
class Entry
{
    public:
        //    private:
        Entry<level-1> left;
        Entry<level-1> right;
    public:

        Entry<level>()
        {
            ;
        }

        void print_positions(FILE *fd)
        {
            left.print_positions(fd);
            fprintf(fd, ", ");
            right.print_positions(fd);
        }

        void gen(Entry<level-1> *left, Entry<level-1> *right, const unsigned int list)
        {
            this->left  = *left;
            this->right = *right;
        }

        /* Returns at least 56 merge bits depending on the level in the least significant bits */
        inline uint64_t get_merge_bits(const unsigned int listid, const unsigned int nlists/*,const unsigned int nbits*/) const
        {
            //            uint64_t mask = (1 << nbits) - 1;
            return (this->get_partial_value<uint64_t>(listid, MERGE_STARTBIT(level,nlists) >> 3) >> (MERGE_STARTBIT(level,nlists) & 7));
            // & mask;
        }

        /* Returns PRESORTED bits needed for PRESORTING on this level */
        inline uint64_t get_presort_bits(const unsigned int listid, const unsigned int nlists) const
        {
            uint64_t mask = (1 << PRESORTED) - 1;
            return (this->get_partial_value<uint64_t>(listid, PRESORT_STARTBIT(level,nlists) >> 3) >> (PRESORT_STARTBIT(level,nlists) & 7)) & mask;
        }

        /* List enumeration starts at 0 for each level */
        template <typename T>
            inline T get_partial_value(const unsigned int listid, const unsigned int byte_offset) const
            {
                return(left.get_partial_value<T>(2 * listid, byte_offset) ^ (right.get_partial_value<T>(2 * listid + 1, byte_offset)));
            }
};

template<>
class Entry<0>
{
    public:
        unsigned char positions[5];
    public:

        Entry<0>()
        {
            ;
        }

        void print_positions(FILE *fd)
        {
            uint16_t pos1, pos2, pos3;

            get_positions(&pos1, &pos2, &pos3);

            fprintf(fd, "%u, %u, %u", pos1, pos2, pos3);
        }

        /* 
         * 0    <= pos1 < 2^14 
         * pos1 <  pos2 < 2^14
         * 0    <= pos3 < 2^13 pos3 always describes the position from the halve-block
         */
        void gen(const uint16_t pos1, const uint16_t pos2, const uint16_t pos3)
        {
            uint64_t dummy = pos3;
            dummy <<= 13;
            if(pos2 < (1 << 13))
            {
                dummy ^= pos2;
                dummy <<= 14;
                dummy ^= pos1;
            }
            else
            {
                dummy ^= (((1 << 14) - 1) - pos2);
                dummy <<= 14;
                dummy ^= (((1 << 14) - 1) - pos1);
            }
            //LITTLEENDIAN
            memcpy(this->positions,&dummy,5);
        };

        /* Returns PRESORTED bits needed for PRESORTING on this level */
        inline uint64_t get_presort_bits(const unsigned int listid, const unsigned int nlists) const
        {
            //            const int level = 0;
            //             unsigned int start = STARTBIT(0) - PRESORTED;
            uint64_t mask = (1 << PRESORTED) - 1;
            return (this->get_partial_value<uint64_t>(listid, PRESORT_STARTBIT(0,nlists) >> 3) >> (PRESORT_STARTBIT(0,nlists) & 7)) & mask;
        }

        /* Returns the merge bits depending on the level in the least significant bits */
        inline uint64_t get_merge_bits(const unsigned int listid, const unsigned int nlists/*, const unsigned int nbits*/) const
        {
            //            const int level = 0;
            //            unsigned int mask = (1 << nbits) - 1;
            return (this->get_partial_value<uint64_t>(listid, MERGE_STARTBIT(0,nlists) >> 3) >> (MERGE_STARTBIT(0,nlists) & 7));// & mask;
        }


        inline void get_positions(uint16_t *pos1, uint16_t *pos2, uint16_t *pos3) const
        {
            uint64_t dummy = 0;
            //LITTLEENDIAN
            memcpy(&dummy,this->positions,5);
            *pos1 = dummy & 0x3fff;
            dummy >>= 14;
            *pos2 = dummy & 0x1fff;
            dummy >>= 13;
            *pos3 = dummy;
            if(*pos1 > *pos2)
            {
                *pos1 = ((1 << 14) - 1) - *pos1;
                *pos2 = ((1 << 14) - 1) - *pos2;
            }
        }

        /* List enumeration starts at 0 for each level */
        template<typename T> 
            inline T get_partial_value(const unsigned int listid, const unsigned int byte_offset) const
            {
                uint16_t pos1 = 0, pos2 = 0, pos3 = 0;
                this->get_positions(&pos1, &pos2, &pos3);
                return get_part_value<T>(pos1, pos2, pos3, listid, byte_offset);
            }
};

#ifndef FINALIZE

/* COMPRESSION CODE STARTING HERE */

template<>
class Entry<2>
//class Entry2
{
    public:
        unsigned char data[LEVEL_2_BYTES];

        void gen(Entry<1> *left, Entry<1> *right, const unsigned int listid)
        {
            //            uint64_t tmp;
            size_t i = 0;
            //            size_t nmm = ((LEVEL_2_BYTES + 15)/16);
            //            UInt128 t[nmm] __attribute__ ((aligned (16)));
            //
            //
            //            for(i = 0; i < nmm; i++)
            //            {
            //                t[i] = left->get_partial_value<UInt128>(listid*2, (ALL_BYTES - LEVEL_2_BYTES) + (i*16)) 
            //                     ^ right->get_partial_value<UInt128>(listid*2+1, (ALL_BYTES - LEVEL_2_BYTES) + (i*16));
            //            }
            //            memcpy(data, t, LEVEL_2_BYTES);


#if (LEVEL_2_BYTES & 16)
            *((uint64_t*)(&data[i])) = left->get_partial_value<uint64_t>(listid*2, (ALL_BYTES- LEVEL_2_BYTES) + i);
            *((uint64_t*)(&data[i])) ^= right->get_partial_value<uint64_t>(listid*2+1, (ALL_BYTES- LEVEL_2_BYTES) + i);
            i += 8;
            *((uint64_t*)(&data[i])) = left->get_partial_value<uint64_t>(listid*2, (ALL_BYTES- LEVEL_2_BYTES) + i);
            *((uint64_t*)(&data[i])) ^= right->get_partial_value<uint64_t>(listid*2+1, (ALL_BYTES- LEVEL_2_BYTES) + i);
            i += 8;
#endif

#if (LEVEL_2_BYTES & 8)
            *((uint64_t*)(&data[i])) = left->get_partial_value<uint64_t>(listid*2, (ALL_BYTES- LEVEL_2_BYTES) + i);
            *((uint64_t*)(&data[i])) ^= right->get_partial_value<uint64_t>(listid*2+1, (ALL_BYTES- LEVEL_2_BYTES) + i);
            i += 8;
#endif

#if (LEVEL_2_BYTES & 4)
            *((uint32_t*)(&data[i])) = left->get_partial_value<uint32_t>(listid*2, (ALL_BYTES- LEVEL_2_BYTES) + i);
            *((uint32_t*)(&data[i])) ^= right->get_partial_value<uint32_t>(listid*2+1, (ALL_BYTES- LEVEL_2_BYTES) + i);
            i += 4;
#endif

#if (LEVEL_2_BYTES & 2)
            *((uint16_t*)(&data[i])) = left->get_partial_value<uint16_t>(listid*2, (ALL_BYTES- LEVEL_2_BYTES) + i);
            *((uint16_t*)(&data[i])) ^= right->get_partial_value<uint16_t>(listid*2+1, (ALL_BYTES- LEVEL_2_BYTES) + i);
            i += 2;
#endif

#if (LEVEL_2_BYTES & 1)
            *((uint8_t*)(&data[i])) = left->get_partial_value<uint8_t>(listid*2, (ALL_BYTES- LEVEL_2_BYTES) + i);
            *((uint8_t*)(&data[i])) ^= right->get_partial_value<uint8_t>(listid*2+1, (ALL_BYTES- LEVEL_2_BYTES) + i);
            i += 1;
#endif


            //            tmp = left->get_partial_value<uint64_t>(listid*2, 8);
            //            tmp ^= right->get_partial_value<uint64_t>(listid*2+1, 8);

            //            memcpy(&data[8], &tmp, 7);

            //            *((uint64_t*)(&data[8])) = left->get_partial_value<uint64_t>(listid*2, 17);
            //            *((uint64_t*)(&data[8])) ^= right->get_partial_value<uint64_t>(listid*2+1, 17);

            //            data[15] = tmp;            
        }

        void print_positions(FILE *fd)
        {
            fprintf(fd, "Using value-only representation: cannot reconstruct positions; ");
        }

        // Returns PRESORTED bits needed for PRESORTING on this level 
        inline uint64_t get_presort_bits(const unsigned int listid, const unsigned int nlists) const
        {
            //            const int level = 2;
            //             unsigned int start = STARTBIT(2) - PRESORTED;
            uint64_t mask = (1 << PRESORTED) - 1;
            return (get_partial_value<uint64_t>(listid, PRESORT_STARTBIT(2,nlists) >> 3) >> (PRESORT_STARTBIT(2,nlists) & 7)) & mask;
        }

        // Returns the merge bits depending on the level in the least significant bits 
        inline uint64_t get_merge_bits(const unsigned int listid, const unsigned int nlists) const
        {
            //            const int level = 2;
            //            unsigned int mask = (1 << nbits) - 1;
            return (get_partial_value<uint64_t>(listid, MERGE_STARTBIT(2,nlists) >> 3) >> (MERGE_STARTBIT(2,nlists) & 7));// & mask;
        }

        // List enumeration starts at 0 for each level 
        template<typename T> 
            inline T get_partial_value(const unsigned int listid, const unsigned int byte_offset) const
            {
                if(byte_offset >= ALL_BYTES - LEVEL_2_BYTES)
                    return *((T*)&data[byte_offset - (ALL_BYTES - LEVEL_2_BYTES)]);
                /* PETER'S VERSION, PLEASE CHECK! */
                else
                {
                    T ret;
                    int i = ALL_BYTES - LEVEL_2_BYTES;
                    unsigned int j = 0;
                    while(i<0 && j < sizeof(T))
                    {
                        j[(unsigned char*)&ret] = 0;
                        ++i;
                        ++j;
                    }
                    i = 0;
                    while(j<sizeof(T) && i < LEVEL_2_BYTES)
                    {
                        j[(unsigned char*)&ret] = data[i];
                        ++i;
                        ++j;
                    }
                    while(j<sizeof(T))
                    {
                        j[(unsigned char*)&ret] = 0;
                        ++j;
                    }
                    return ret;
                }
                //            printf("ALL_BYTES: %u   LEVEL_2_BYTES %u\n", ALL_BYTES, LEVEL_2_BYTES);
                //            printf("ALL_BITS: %u   LEVEL_2_BITS %u\n", ALL_BITS, LEVEL_2_BITS);
            }
};

template<>
class Entry<3>
//class Entry3
{
    public:
        unsigned char data[LEVEL_3_BYTES];

        void gen(Entry<2> *left, Entry<2> *right, const unsigned int list = 0)
            //        void gen(Entry2 *left, Entry2 *right, const unsigned int list = 0)
        {
            //            uint8_t tmp = data[15];

            //            *((uint64_t*)(&data[0])) = *((uint64_t*)(&left->data[5]));
            //            *((uint64_t*)(&data[0])) ^= *((uint64_t*)(&right->data[5]));

            //            *((uint16_t*)(&data[8])) = *((uint16_t*)(&left->data[13]));
            //            *((uint16_t*)(&data[8])) ^= *((uint16_t*)(&right->data[13]));
            int i = 0;

#if (LEVEL_3_BYTES & 16)
            *((uint64_t*)(&data[i])) = *((uint64_t*)(&left->data[LEVEL_2_BYTES - LEVEL_3_BYTES + i]));
            *((uint64_t*)(&data[i])) ^= *((uint64_t*)(&right->data[LEVEL_2_BYTES - LEVEL_3_BYTES + i]));
            i += 8;
            *((uint64_t*)(&data[i])) = *((uint64_t*)(&left->data[LEVEL_2_BYTES - LEVEL_3_BYTES + i]));
            *((uint64_t*)(&data[i])) ^= *((uint64_t*)(&right->data[LEVEL_2_BYTES - LEVEL_3_BYTES + i]));
            i += 8;
#endif

#if (LEVEL_3_BYTES & 8)
            *((uint64_t*)(&data[i])) = *((uint64_t*)(&left->data[LEVEL_2_BYTES - LEVEL_3_BYTES + i]));
            *((uint64_t*)(&data[i])) ^= *((uint64_t*)(&right->data[LEVEL_2_BYTES - LEVEL_3_BYTES + i]));
            i += 8;
#endif

#if (LEVEL_3_BYTES & 4)
            *((uint32_t*)(&data[i])) = *((uint32_t*)(&left->data[LEVEL_2_BYTES - LEVEL_3_BYTES + i]));
            *((uint32_t*)(&data[i])) ^= *((uint32_t*)(&right->data[LEVEL_2_BYTES - LEVEL_3_BYTES + i]));
            i += 4;
#endif

#if (LEVEL_3_BYTES & 2)
            *((uint16_t*)(&data[i])) = *((uint16_t*)(&left->data[LEVEL_2_BYTES - LEVEL_3_BYTES + i]));
            *((uint16_t*)(&data[i])) ^= *((uint16_t*)(&right->data[LEVEL_2_BYTES - LEVEL_3_BYTES + i]));
            i += 2;
#endif

#if (LEVEL_3_BYTES & 1)
            *((uint8_t*)(&data[i])) = *((uint8_t*)(&left->data[LEVEL_2_BYTES - LEVEL_3_BYTES + i]));
            *((uint8_t*)(&data[i])) ^= *((uint8_t*)(&right->data[LEVEL_2_BYTES - LEVEL_3_BYTES + i]));
            i += 1;
#endif


            //            tmp = left->get_partial_value<uint64_t>(listid*2, 8);
            //            tmp ^= right->get_partial_value<uint64_t>(listid*2+1, 8);

            //            memcpy(&data[8], &tmp, 7);

            //            *((uint64_t*)(&data[8])) = left->get_partial_value<uint64_t>(listid*2, 17);
            //            *((uint64_t*)(&data[8])) ^= right->get_partial_value<uint64_t>(listid*2+1, 17);

            //            data[15] = tmp;            
        }

        void print_positions(FILE *fd)
        {
            fprintf(fd, "Using value-only representation: cannot reconstruct positions; ");
        }

        // Returns PRESORTED bits needed for PRESORTING on this level 
        inline uint64_t get_presort_bits(const unsigned int listid, const unsigned int nlists) const
        {
            //            const int level = 3;
            //             unsigned int start = STARTBIT(3) - PRESORTED;
            uint64_t mask = (1 << PRESORTED) - 1;
            return (get_partial_value<uint64_t>(listid, PRESORT_STARTBIT(3,nlists) >> 3) >> (PRESORT_STARTBIT(3,nlists) & 7)) & mask;
        }

        // Returns the merge bits depending on the level in the least significant bits 
        inline uint64_t get_merge_bits(const unsigned int listid, const unsigned int nlists) const
        {
            //            const int level = 3;
            //            unsigned int mask = (1 << nbits) - 1;
            return (get_partial_value<uint64_t>(listid, MERGE_STARTBIT(3,nlists) >> 3) >> (MERGE_STARTBIT(3,nlists) & 7));// & mask;
        }

        // List enumeration starts at 0 for each level 
        template<typename T> 
            inline T get_partial_value(const unsigned int listid, const unsigned int byte_offset) const
            { 
                T ret = 0;

                if (byte_offset < (ALL_BYTES - LEVEL_3_BYTES))
                {
                    for (int i = 0; i < LEVEL_3_BYTES; i++)
                        ((uint8_t*)&ret)[i + (ALL_BYTES - LEVEL_3_BYTES) - byte_offset] = data[i];
                }
                else
                {
                    ret = *((T*)&data[byte_offset - (ALL_BYTES - LEVEL_3_BYTES)]);
                }
                return ret;

                /* PETER'S VERSION
                   if(byte_offset >= ALL_BYTES - LEVEL_3_BYTES)
                   return *((T*)&data[byte_offset - (ALL_BYTES - LEVEL_3_BYTES)]);
                   else
                   {
                   T ret;
                   int i = ALL_BYTES - LEVEL_3_BYTES;
                   unsigned int j = 0;
                   while(i<0 && j < sizeof(T))
                   {
                   j[(unsigned char*)&ret] = 0;
                   ++i;
                   ++j;
                   }
                   i = 0;
                   while(j<sizeof(T) && i < LEVEL_3_BYTES)
                   {
                   j[(unsigned char*)&ret] = data[i];
                   ++i;
                   ++j;
                   }
                   while(j<sizeof(T))
                   {
                   j[(unsigned char*)&ret] = 0;
                   ++j;
                   }
                   return ret;
                   } */
            }
};

#endif

/* COMPRESSION CODE ENDING HERE */

template <uint8_t level>
class NetEntry
{
    public:

        Entry<level> e;
        uint16_t presort_bits;

        void gen(Entry<level-1> *left, Entry<level-1> *right, const unsigned int listid, const unsigned int nlists)
        {
            uint64_t mask = (1 << PRESORTED) - 1;

            e.gen(left, right, listid);

            presort_bits = ((left->template get_partial_value<uint16_t>(2*listid, 
                            PRESORT_STARTBIT(level,nlists) >> 3) >> (PRESORT_STARTBIT(level,nlists) & 7)) ^
                    (right->template get_partial_value<uint16_t>(2*listid + 1,
                                                                 PRESORT_STARTBIT(level,nlists) >> 3) >> (PRESORT_STARTBIT(level,nlists) & 7))) & mask;
        }

        uint16_t get_presort_bits(int list)
        {
            return presort_bits;
        }

        void print_positions(FILE *fd)
        {
            e.print_positions(fd);
        }

        inline uint64_t get_merge_bits(const unsigned int listid, const unsigned int nlists) const
        {
            return e.get_merge_bits(listid, nlists);
        }

        template<typename T> 
            inline T get_partial_value(const unsigned int listid, const unsigned int byte_offset) const
            {
                return e.template get_partial_value<T>(listid, byte_offset);
            }

        Entry<level> *getEntry()
        {
            return &e;
        }

};

#endif

