#ifndef __FSBDAY_H_
#define __FSBDAY_H_

#include <pthread.h>

class Singleton
{
    private:
        static volatile int global_counter;

        static pthread_barrier_t *pbarrier;
        static int num_barrier_threads;

    public:
        Singleton()
        {
            num_barrier_threads++;

            if (pbarrier == NULL)
            { 
                pbarrier = (pthread_barrier_t*)malloc(sizeof(pthread_barrier_t));

                pthread_barrier_init(pbarrier, NULL, num_barrier_threads);
            }
            else
            {
                pthread_barrier_destroy(pbarrier);

                pthread_barrier_init(pbarrier, NULL, num_barrier_threads);
            }
        }

        ~Singleton()
        {
            if (pbarrier != NULL)
            {
                pthread_barrier_destroy(pbarrier);
                free(pbarrier);
            }
        }

        void barrier()
        {
            pthread_barrier_wait(pbarrier);
        }

        bool enter()
        {
            barrier();

            return __sync_bool_compare_and_swap(&Singleton::global_counter, 0, 1);
        }

        void leave()
        {
            barrier();

            this->global_counter = 0;
        }
};

int Singleton::num_barrier_threads = 0;
pthread_barrier_t *Singleton::pbarrier = NULL;
volatile int Singleton::global_counter = 0;


template <class T>
class FSBdayTask : public T, public Singleton
{
    void enter_round()
    {
        if (Singleton::enter())
        {
/*            INFO_DEBUG(VERBOSITY > 0,"Number of elements in my list 0: %lu\n", 
                    alesys->get_list_size(0));
            INFO_DEBUG(VERBOSITY > 0,"Number of elements in my list 1: %lu\n", 
                    alesys->get_list_size(1));

            alesys->swap_in_out();*/

            MPI_Barrier(MPI_COMM_WORLD);
        }
        Singleton::leave();
    }

    private:
    Argv *argv;

    public:
    virtual void do_job(int level, int tid) = 0;
    
    virtual void init() 
    { 
        ;
    }

    Argv *getArgv()
    {
        return argv;
    }

    void run(void *argv, int tid)
    {
        int round;

        this->argv = (Argv*)argv;

        init();

        for (round = this->argv->startlevel; !this->argv->is_final_level(round); round++)
        {
            enter_round();
            do_job(round, tid);
        }

        Singleton::barrier();
    }
};

#endif

