#!/bin/bash

HOSTFILE=./hosts
FSBDAY=./fsbday
MPDBOOT="mpdboot -n 8 -f $HOSTFILE"
MPIRUN="mpirun -l -machinefile $HOSTFILE -np 8"
CURR_PATH=$(pwd)

if [ -n "$1" ];then
    OUTFILE=$1
    echo "" > $OUTFILE
fi

fsbstart(){
    STOP=$(cat ./stopfile)
    if [ "$STOP" = "stop" ];then
        echo "requested stop; exiting"
        exit 0  
    fi

    while read HOST; do
        ssh $HOST "cp $CURR_PATH/$FSBDAY $CURR_PATH/$FSBDAY.run" < /dev/null
    done < ./hosts

    if [ -n "$OUTFILE" ];then
        date >> $OUTFILE
        echo "Running $1" >> $OUTFILE
        $MPIRUN $FSBDAY.run $1 2>> $OUTFILE
    else
        echo "Running $1"
        date
        $MPIRUN $FSBDAY.run $1
    fi
}

SUCCESS=0
C=0

$MPDBOOT

date
fsbstart "--startlevel 0 --startlist 0 --nlists 8 --persist"

while read HOST; do
    ssh $HOST "cd /var/fsbday/ && cp -r ale ale_left_half_tree && chgrp fsbday ale_left_half_tree && chmod g+rwx ale_left_half_tree" < /dev/null
done < ./hosts

date
fsbstart "--startlevel 0 --startlist 8 --nlists 4 --persist"

while read HOST; do
    ssh $HOST "cd /var/fsbday/ && cp -r ale ale_third_quarter_tree && chgrp fsbday ale_third_quarter_tree && chmod g+rwx ale_third_quarter_tree" < /dev/null
done < ./hosts

ROUND=0

while [ $C -lt 4096 ] && [ $SUCCESS -eq 0 ];do

  C0=$(($C % 8))
  C1=$((($C / 8) % 8))
  C2=$((($C / 64) % 8))
  C3=$((($C / 512) % 8))
  C=$(($C+1))
  while [ $(($C0 ^ $C1 ^ $C2 ^ $C3)) -ne 0 ];do
    C0=$(($C % 8))
    C1=$((($C / 8) % 8))
    C2=$((($C / 64) % 8))
    C3=$((($C / 512) % 8))
    C=$(($C+1))
  done

  /bin/date
  echo "$ROUND Trying clamping constants: $C0 $C1 $C2 $C3 (C = $((C-1)))"
  fsbstart "--startlevel 0 --startlist 12 --nlists 4 $C0 $C1 $C2 $C3"
  fsbstart "--startlevel 2 --startlist 2  --nlists 2"
  fsbstart "--startlevel 3 --startlist 0  --nlists 2"
  SUCCESS=$?

  /bin/date
  if [ $SUCCESS -eq 0 ];then
    echo "Tried clamping constants: $C0 $C1 $C2 $C3 (C = $((C-1))) -- no success"
  else
    echo "Success with clamping constants: $C0 $C1 $C2 $C3 (C = $((C-1)))"
  fi

  ROUND=$(($ROUND+1))

done
