#ifndef DEBUG_H
#define DEBUG_H

#include <signal.h>
#include <execinfo.h>
#include <sys/time.h>
#include <stdlib.h>

#ifdef DEBUG

#define HALT 1
#define NOHALT 0

extern struct timeval start, end;

#define ERROR_DEBUG(condition, halt, format, args...) \
    if (condition) { \
        gettimeofday(&end, NULL);\
\
        uint64_t sec, min, h, d;\
\
        sec = end.tv_sec - start.tv_sec;\
        min = sec / 60;\
        sec = sec % 60;\
        \
        h = min / 60;\
        min = min % 60;\
        \
        d = h / 24;\
        h = h % 24;\
\
        fprintf(stderr, "%s:%u(%s) %02lu:%02lu:%02lu:%02lu - "format, __FILE__, __LINE__, __FUNCTION__, d, h, min, sec, ##args); \
        fflush(stderr);\
        if (halt == HALT) \
        {\
            void *array[10];\
            size_t size;\
            size = backtrace(array, 10);\
            backtrace_symbols_fd(array, size, 2);\
            raise(1); \
            exit(-1);\
        }\
    }

#define INFO_DEBUG(condition, format, args...) \
    if(condition) \
    { \
        gettimeofday(&end, NULL);\
\
        uint64_t sec, min, h, d;\
\
        sec = end.tv_sec - start.tv_sec;\
        min = sec / 60;\
        sec = sec % 60;\
        \
        h = min / 60;\
        min = min % 60;\
        \
        d = h / 24;\
        h = h % 24;\
\
        fprintf(stderr, "%s:%u(%s) %02lu:%02lu:%02lu:%02lu - "format, __FILE__, __LINE__, __FUNCTION__, d, h,   min, sec, ##args); \
        fflush(stderr); \
    }

#else

#define ERROR_DEBUG(condition, halt, format, args...)
#define INFO_DEBUG(condition, format, args...)

#endif

#endif //DEBUG_H

