#ifndef LIST_H
#define LIST_H

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <time.h>
#include <pthread.h>
#include <sys/mman.h>

#include "params.h"
#include "entry.h"
#include "include/debug.h"

template <unsigned level>
class List
{
  public:
    Entry<level> *data;
    uint64_t len; // number of stored elements
    uint8_t id;

  public:

    List(uint8_t id)
    {
      this->id = id;
      this->len = 0;
    }

    ~List()
    {
      ;
    }

    void gen(Entry<level> *data, uint64_t len)
    {
      this->data = data;
      this->len = len;
    }

    void sort()
    {
      do_sort(data, len, 
          (level + 1) * SORT_LEVEL_RANGE + CLIPPING,
          level       * SORT_LEVEL_RANGE + CLIPPING  + PRESORTED + (level + 1),
          SORT_BLOCK_SIZE);
    }

  private:

    // loc_data: pointer to beginning of list
    // nr: nimber of elements to sort beginning at loc_data
    // block_start: start-bit-number for range of the value to sort
    // block_end: end-bit-number
    // block_size: number of bits per block
    void do_sort(Entry<level> *loc_data, uint32_t nr, uint8_t block_start, uint8_t block_end, uint8_t block_size)
    {
      uint32_t count[1 << block_size];
      Entry<level> *pile[1 << block_size];

      Entry<level> *ep, *pl;

      Entry<level> r, t;

      int i, c;

      uint8_t block = (block_start - block_end) > block_size ? 
        block_size : 
        (block_start - block_end);

      uint8_t shift = (block_start - (block_end & ~7) - block);

      uint64_t mask = 0xffffffffffffffffUL >> (64 - block);

      for (i = 0; i < 1 << block; i++)
        count[i] = 0;

      // count
      for (ep = loc_data; ep < loc_data + nr; ep += 1)
      {
        count[(uint16_t)((ep-> template get_partial_value<uint64_t>
              (id, block_end >> 3)) >> shift) & mask]++; 
      }

      pl = loc_data;
      // compute piles
      for (i = 0; i < 1 << block; i++)
        pile[i] = pl += count[i];

      c = 0;

      // sort
      for (ep = loc_data; ep < loc_data + nr; ep += count[c])
      {
        r = *ep;
        while((pile[c = (uint16_t)((r.template get_partial_value<uint64_t>
                  (id, block_end >> 3) >> shift) & mask)] -= 1) > ep)
        {
          t = *pile[c];
          *pile[c] = r;
          r = t;
        }
        *ep = r;
      }

      // recurse
      if (block_start-block > block_end)
      {
        pl = loc_data;
        for (i = 0; i < 1 << block; i++)
        {
          if (count[i] > 1)
          {
            do_sort(pl, count[i],
                block_start-block, block_end, block_size);
          }
          pl += count[i];
        }
      }

    }

};

#endif
