#include "fsbday.h"
#include "task.h"

class Merge : public FSBdayTask<CrossTask<Part*, NetData*> >
{
    public:

        template<uint8_t level>
            int merge(const uint32_t partnr, part_t *part0, part_t *part1, uint64_t nelems0, uint64_t nelems1, NetData **net_data, Source<NetData*> *source, SinkList<NetData*> *sink)
            {
                uint64_t cxor0;
                uint64_t cxor1;

                uint64_t nelems1_sav, nelems_gen;
                uint8_t node;
                NetEntry<level+1> *e;

                Entry<level> *t0 = (Entry<level> *) part0;
                Entry<level> *t1 = (Entry<level> *) part1;

                Entry<level> *l0; 
                Entry<level> *l1;

                nelems_gen = 0;

                Argv *argv = getArgv();
                uint8_t list0 = argv->get_list0(level, rank);
                unsigned int nlists = argv->get_nlists(level);
                unsigned long nodemask = MIN((4*NNODES/nlists),NNODES) - 1;
                unsigned int node_offset = (list0 & ~nodemask) % NNODES;

                INFO_DEBUG(VERBOSITY > 1, "node_offset: %u  list0/2: %u  nlists: %u  nodemask: %lu\n", node_offset, list0/2, nlists, nodemask);

                ERROR_DEBUG((nelems0 * sizeof(Entry<level>)) > sizeof(part_t), NOHALT,
                        "Got more elemts to merge than space in part! part: %u list: %u nelems: %lu\n",
                        0, list0, nelems0);

                ERROR_DEBUG((nelems1 * sizeof(Entry<level>)) > sizeof(part_t), NOHALT,
                        "Got more elemts to merge than space in part! part: %u list: %u nelems: %lu\n",
                        1, list0 + 1, nelems1);

                cxor0 = t0->get_merge_bits(list0, nlists);
                cxor1 = t1->get_merge_bits(list0+1, nlists);


                while((nelems0 > 0) && (nelems1 > 0))
                {
                    INFO_DEBUG(3, "cxor0: %lx  cxor1: %lx\n", cxor0, cxor1);
                    if((cxor0 & CMP_MASK(level,nlists)) == (cxor1 & CMP_MASK(level,nlists)))
                    {
                        l0 = t0;
                        l1 = t1;

                        nelems1_sav = nelems1;
                        do
                        {
#if FINALIZE
                            if (argv->is_final_level(level + 1))
                            {
                                NetEntry<level+1> dummy;
                                uint64_t cxx;
                                uint64_t rxx;
                                uint64_t xrbits = REMAINING_BITS - NODEBITS(nlists/2) - PRESORTED;
                                uint64_t rmask = ((1UL << xrbits) - 1) << (64 - xrbits);
                                node = (((cxor0^cxor1) >> CMP_NBITS(level,nlists)) & nodemask) + node_offset;
                                dummy.gen(l0, l1, list0/2, nlists/2);
                                cxx = dummy.get_merge_bits(list0/2, nlists/2);
                                rxx = dummy.template get_partial_value<uint64_t>(list0/2,  ALL_BYTES - sizeof(uint64_t));
                                if((cxx & CMP_MASK(level+1,nlists/2)) == (argv->get_cmp() & CMP_MASK(level+1,nlists/2)) \
                                        && (rxx & rmask) == (argv->get_remaining() & rmask) \
                                        && dummy.get_presort_bits(list0/2) == argv->get_part() \
                                        && node == argv->get_node())
                                {
                                    fprintf(stderr, "POSITIONS\n");
                                    dummy.print_positions(stderr);
                                }
                            }
#else
                            if ((level + 1) == MAX_LEVEL)
                            {
                                uint64_t xxx0;
                                uint64_t xxx1;

                                xxx0 = l0->template get_partial_value<uint64_t>(list0,  ALL_BYTES - sizeof(uint64_t));
                                xxx1 = l1->template get_partial_value<uint64_t>(list0+1,  ALL_BYTES - sizeof(uint64_t));
                                
                                if ( xxx0 == xxx1)
                                {
                                    fprintf(stderr, "\n");
                                    fprintf(stderr, "Result: %lx %lx  Part: %x\n", cxor0, 
                                            l0->template get_partial_value<uint64_t>(list0,  ALL_BYTES - sizeof(uint64_t)), partnr);
                                    success = 1;
                                }
                            }
#endif
                            else
                            {
                                node = (((cxor0^cxor1) >> CMP_NBITS(level,nlists)) & nodemask) + node_offset;

                                ERROR_DEBUG(node > NNODES, NOHALT, 
                                        "Value of node too high: %u CMP_NBITS: %u nodemask: %lu, node_offset: %u\n", 
                                        node, CMP_NBITS(level,nlists), nodemask, node_offset);

                                e = net_data[node]->get_slot<NetEntry<level+1> >();
                                if(e == NULL)
                                {
                                    INFO_DEBUG(VERBOSITY > 5, "NetData full, sending merged data to node: %u\n", node);
                                    sink->getSink(node%NNODES)->put(net_data[node]);
                                    net_data[node] = source->get();

                                    net_data[node]->reset(list0/2);

                                    e = net_data[node]->get_slot<NetEntry<level+1> >();
                                }
                                e->gen(l0, l1, list0/2, nlists/2);
                                nelems_gen++;
                            }

                            l1++;
                            nelems1--;

                            cxor1 = l1->get_merge_bits(list0+1, nlists);

                            if((cxor0 & CMP_MASK(level, nlists)) != (cxor1 & CMP_MASK(level, nlists)))
                            {
                                l1 = t1;
                                nelems1 = nelems1_sav;
                                l0++;
                                nelems0--;
                                cxor0 = l0->get_merge_bits(list0, nlists);
                                cxor1 = l1->get_merge_bits(list0+1, nlists);
                            }
                        }
                        while(((cxor0 & CMP_MASK(level, nlists)) == (cxor1 & CMP_MASK(level, nlists))) && (nelems0 > 0) && (nelems1 > 0));

                        t0 = l0;
                    }
                    if((nelems0 > 0) && (nelems1 > 0))
                    {
                        if((cxor0 & CMP_MASK(level, nlists)) > (cxor1 & CMP_MASK(level, nlists))) 
                        {
                            t1++;
                            cxor1 = t1->get_merge_bits(list0+1, nlists);
                            nelems1--;
                        }
                        if((cxor0 & CMP_MASK(level, nlists)) < (cxor1 & CMP_MASK(level, nlists))) 
                        {
                            t0++;
                            cxor0 = t0->get_merge_bits(list0, nlists);
                            nelems0--;
                        }
                    }
                }

                return nelems_gen;
            } 

        template <uint8_t level> void prepare_merge(uint32_t id, Source<NetData*> *source, SinkList<NetData*> *sink)
            {
                uint32_t i, node;

                Part *data0, *data1;

                uint64_t nelems_gen;

                NetData *net_data[NNODES];

                Argv *argv = getArgv();

                unsigned int list0 = argv->get_list0(level, rank);
                unsigned int nlists = argv->get_nlists(level);

                INFO_DEBUG(VERBOSITY >= 0, "Called with list0 == %u, producing values for list %u, level %i, nlists %u\n", list0, list0/2, level+1, nlists/2);
                INFO_DEBUG(VERBOSITY >= 0, "Starting merge with LEVEL_STARTBIT = %u, PRESORT_STARTBIT = %u, MERGE_STARTBIT = %u\n",
                        LEVEL_STARTBIT(level), PRESORT_STARTBIT(level,nlists), MERGE_STARTBIT(level,nlists)); 

                for (node = 0; node < NNODES; node++)
                {
                    net_data[node] = source->get();
                    net_data[node]->reset(list0/2);
                }

                for (i = 0; i < NPARTS; i++)
                {
                    data0 = getSourceA()->getSource(0)->get();
                    data1 = getSourceA()->getSource(1)->get();

                    INFO_DEBUG(VERBOSITY > 0, "merge level: %u  part: %u  list0: %u list1: %u  nelems0: %lu nelems1: %lu\n", 
                            level, i, list0, list0+1, data0->nelems, data1->nelems);

                    nelems_gen = merge<level>(i, data0->data, data1->data, 
                            data0->nelems, data1->nelems, 
                            net_data, source, sink);

                    INFO_DEBUG(VERBOSITY > 1, "genereated elements: %lu\n", nelems_gen);

                    getSinkA()->getSink(0)->put(data0);
                    getSinkA()->getSink(1)->put(data1);                    

                    MPI_Barrier(MPI_COMM_WORLD); // avoid getting out of sync; we might get an deadlock otherwise!
                }

                for (node = 0; node < NNODES; node++)
                {
                    INFO_DEBUG(VERBOSITY > 1, "End of merge, sending merged data to node: %u\n", node);

                    net_data[node]->set_last_of_level();
                    sink->getSink(node)->put(net_data[node]);
                }

                INFO_DEBUG(VERBOSITY > 1, "thread %u finished level %u\n", id, level);
            }

        void do_job(int level, int tid)
        {

            switch (level)
            {
                case 0: prepare_merge<0>(tid, getSourceB(), getSinkB());
                        break;
                case 1: 
                        prepare_merge<1>(tid, getSourceB(), getSinkB());
                        break;
                case 2:
                        prepare_merge<2>(tid, getSourceB(), getSinkB());
                        break;
                case 3:
                        prepare_merge<3>(tid, getSourceB(), getSinkB());
                        break;
            }

        }

        ~Merge()
        {
            INFO_DEBUG(VERBOSITY > 0, "finished sort and merge\n");
        }
};

