#ifndef PARAMS_H
#define PARAMS_H

#include <inttypes.h>

/* Output verbosity for (testing) programs */
#define VERBOSITY 1                                   // Debug output verbosity

/* Parameters concerning the problem instance */
#define MATRIX_COLUMNS (3 * (1 << 17))                // Number of columns of the matrix
#define COLUMN_BYTES 24                               // Number of bytes per column
#define W 24                                          // Parameter w in FSB_{48}
#define MATRIX_BYTES (COLUMN_BYTES * MATRIX_COLUMNS)  // Size of the matrix in bytes
#define BLOCK_COLUMNS (MATRIX_COLUMNS / W)            // Number of columns per block
#define MAX_LEVEL 4                                   // Depth of tree

/* Parameters describing the cluster configuration */
#define NNODES 8                                      // Number of cluster nodes

/* Parameters concerning hardware of each cluster node */
#define IODEVICE "/dev/sda1"                          // Device used to store all data
#define DEVICE_BYTES 728545950720UL                   // Number of bytes this device can store

#define ALE_PATH "/var/fsbday/ale/"                   // Path to store persistent information of the AleSystem

/* Parameters concerning the algorithm */
#define SORT_BLOCK_SIZE 9                             // Number of bits used in one run of radix sort
#define SORT_LEVEL_RANGE 37                           // Total number of bits clamped on each level
#define CLIPPING 3                                    // number of bits that are clamped during generation
#define MAXCLAMP ((1 << CLIPPING)-1)                  // Maximal clamping constant during list generation
#define PRESORTED 9                                   // Number of bits used for presorting

#define ALL_BITS 192                                  // Number of bits in a level-0 entry before *any* clamping

#define NPARTS (1 << PRESORTED)                       // Number of parts, must be a power of 2 
#define PARTMASK (NPARTS - 1)                         // Mask used to find the bits identifying the part 

#define REMAINING_BITS (ALL_BITS - CLIPPING - (MAX_LEVEL-1)*SORT_LEVEL_RANGE) // Remaining bits in the two lists on the 

#endif
