#ifndef PRESORT_H
#define PRESORT_H

#include "fsbday.h"
#include "task.h"
#include "entry.h"
#include "consts.h"

class Presort : public FSBdayTask<BasicCrossTask<Source<NetData*>, Sink<NetData*>, Source<Ale*>, Sink<Ale*> > >
{
    // list : number of list, range 0-15
    void gen_elements(uint32_t list, Endpoint<Ale*> *store, uint16_t clamp)
    {
        int i, j, k, part;
        uint32_t cxor;
        Argv *args = getArgv();

        uint16_t nc = (NNODES/args->startnlists + CLIPPING); // we know that startlevel == 0, so we can use startnlists
        uint16_t clampmask = (1 << nc)-1;
        uint16_t clampvalue = ((rank % (1 << (NNODES/args->startnlists))) << CLIPPING) ^ (args->clamp[list - args->startlist]);

        Ale *cur_ales[NPARTS];
        uint32_t ale_pos[NPARTS];

        Entry<0> *t;

        INFO_DEBUG(1, "rank: %u  gen list: %u  nc: %u  clampmask: %u  clampvalue: %u\n", rank, list, nc, clampmask, clampvalue);

        for (i = 0; i < NPARTS; i++)
        {
            cur_ales[i] = store->source->get();
            ale_pos[i] = 0;
            cur_ales[i]->is_last = false;
        }

        INFO_DEBUG(1, "start generate\n");

        for(i=0;i<16384;i++)
        {
            INFO_DEBUG(VERBOSITY > 6,"%u\n", i);
            for(j=i+1;j<16384;j++)
            {
                for(k=0;k<8192;k++)
                {
                    cxor = get_part_value<uint32_t>(i,j,k,list,0);
                    if((cxor & clampmask) == clampvalue)
                    {
                        part = (cxor >> nc) & PARTMASK;

                        t = (Entry<0> *)(((unsigned char *)(cur_ales[part]->data)) + ale_pos[part]);
                        t->gen(i,j,k);

                        ale_pos[part] += sizeof(Entry<0>);
                        if((ale_pos[part] + sizeof(Entry<0>)) > sizeof(ale_t))
                        {
                            cur_ales[part]->part = part;
                            cur_ales[part]->nelems = ale_pos[part]/sizeof(Entry<0>);
                            cur_ales[part]->list = list % LISTFRACTIONS_PER_NODE;
                            store->sink->put(cur_ales[part]);
                            cur_ales[part] = store->source->get();
                            ale_pos[part] = 0;
                            cur_ales[part]->is_last = false;
                        }
                    }
                }
            }
        }

        /* Writing remaining (partially filled) ales */
        for(part=0;part<NPARTS;part++)
        {
            cur_ales[part]->part = part;
            cur_ales[part]->nelems = ale_pos[part]/sizeof(Entry<0>);
            cur_ales[part]->is_last = true;
            cur_ales[part]->list = list % LISTFRACTIONS_PER_NODE;
            store->sink->put(cur_ales[part]);
        }

        INFO_DEBUG(1, "rank: %u finished generation of list %u\n", rank, list);
    }

        template<uint8_t level>
            void presort(uint32_t tid, 
                    Endpoint<NetData*> *net, Endpoint<Ale*> *store)
            {
                int reps;

                uint16_t part;
                uint8_t list;
                uint64_t elements[NPARTS];
                uint8_t alelist = tid;
                part = 0;
                list = 0;

                NetData *net_data;

                Ale *hdd_buffer[NPARTS];

                NetEntry<level> *entry;

                INFO_DEBUG(VERBOSITY > 1, "thread %u start level %i\n", tid, level);

                INFO_DEBUG(1, "presort level %i\n", level);


                for (part = 0; part < NPARTS; part++)
                {
                    hdd_buffer[part] = store->source->get();
                    elements[part] = 0;
                }

                reps = 0;

                do 
                {
                    net_data = net->source->get();

                    ERROR_DEBUG((net_data->get_nelems<NetEntry<level> >() * sizeof(Entry<level>)) > sizeof(part_t), NOHALT,
                            "Got more elemts to merge than space in part! part: %u list: %u nelems: %lu\n",
                            0, list, net_data->get_nelems<NetEntry<level> >());

                    entry = (NetEntry<level>*)(net_data->get_data());

                    list = net_data->get_listid();

                    for (uint64_t i = 0; i < net_data->get_nelems<NetEntry<level> >(); i++)
                    {
                        part =  entry[i].get_presort_bits(list);

                        ERROR_DEBUG((part > 512), HALT, "bad part! part: %u  i: %lu\n", part, i);

                        ((Entry<level>*)(hdd_buffer[part]->data))[elements[part]] = *entry[i].getEntry();

                        elements[part]++;

                        if ((elements[part] + 1) *  sizeof(Entry<level>) > sizeof(ale_t))
                        {
                            hdd_buffer[part]->part = part;
                            hdd_buffer[part]->list = alelist % LISTFRACTIONS_PER_NODE;
                            hdd_buffer[part]->nelems = elements[part];

                            store->sink->put(hdd_buffer[part]);

                            hdd_buffer[part] =  store->source->get();
                            elements[part] = 0;
                        }
                    }

                    if (net_data->is_last_of_level())
                    {
                        reps++;
                        INFO_DEBUG(VERBOSITY > 1, "received %u finished parts of %u...\n", reps, (NNODES / (getArgv()->get_nlists(level) == 1 ? 1 : 2)));
                    }

                    net->sink->put(net_data);

                } while (reps < (NNODES / (getArgv()->get_nlists(level) == 1 ? 1 : 2))); // presort runs with 2 threads, therefore we devide by 2

                // ensure all data is written so that swap_in_out can safely be called on alesystem
                for (part = 0; part < NPARTS; part++)
                {
                    INFO_DEBUG(VERBOSITY > 4, "thread %u on level %u write %lu elements to alelist %u of part %u\n", tid, level,
                            elements[part], alelist % LISTFRACTIONS_PER_NODE, part);

                    hdd_buffer[part]->part = part;
                    hdd_buffer[part]->list = alelist % LISTFRACTIONS_PER_NODE;
                    hdd_buffer[part]->nelems = elements[part];

                    hdd_buffer[part]->is_last = true;

                    store->sink->put(hdd_buffer[part]);
//                    alesys->write_ale(alelist % LISTFRACTIONS_PER_NODE, i, elements[i], hdd_buffer[i], store);
                }

//                while (!store->is_full()) usleep(100);

                INFO_DEBUG(VERBOSITY > 1, "thread %u done level %i\n", tid, level);
            }

    public:

        Presort()
        {
//            avail_ales = new BufferManager<ale_t>(2 * NPARTS);
        }


        void init()
        {
            Endpoint<Ale*> ep(getSourceB(), getSinkB());

            Argv *argv = getArgv();

            uint32_t listid = argv->get_list0(0, rank) + getTID();
            uint16_t clamp = argv->clamp[listid - argv->startlist];

            if (argv->startlevel == 0)
                gen_elements(listid, &ep, clamp); // Last argument is the clamping constant
        }

        void do_job(int level, int tid)
        {
            Endpoint<NetData*> net_ep(getSourceA(), getSinkA());
            Endpoint<Ale*> store_ep(getSourceB(), getSinkB());

            if (getArgv()->is_final_level(level+1)
                    && ((unsigned)tid != (getArgv()->get_list0(level, rank) >> 1) % 2))
            {
                INFO_DEBUG(VERBOSITY > 1, "no work for presort-thread %u; exiting... list0: %u\n", tid, getArgv()->get_list0(level, rank));
                return;
            }

            switch (level+1)
            {
                case 1: presort<1>(tid, &net_ep, &store_ep);
                        break;
                case 2: presort<2>(tid, &net_ep, &store_ep);
                        break;
                case 3: presort<3>(tid, &net_ep, &store_ep);
                        break;
                case 4: presort<4>(tid, &net_ep, &store_ep);
                        break;
            }

        }

        ~Presort()
        {
          ;
        }
};

#endif
