#include "fsbday.h"
#include "list.h"
#include "task.h"

class Sort : public FSBdayTask <PipelineTask<Part*> >
{
    template <uint8_t level>
        void sort(int tid, Source<Part*> *source, Sink<Part*> *sink)
        {
            Part *part;
            int i;

            Argv *argv = getArgv();

            uint8_t listid = argv->get_list0(level, rank) + getTID();

            for (i = 0; i < NPARTS; i++)
            {
                part = source->get();

                ERROR_DEBUG((part->nelems * sizeof(Entry<level>)) > sizeof(part_t), NOHALT,
                        "Got more elemts to sort than space in part! part: %u list: %u nelems: %lu\n",
                        tid, listid, part->nelems);

                INFO_DEBUG(VERBOSITY > 1, "sorting level %u part %u list %u nelems %lu\n", level, i, listid, part->nelems);

                {
                    List<level> l(listid);
                    l.gen((Entry<level> *)part->data, part->nelems);

                    l.sort();
                }

                sink->put(part);
            }
        } 

    void do_job(int level, int tid)
    {
        switch (level)
        {
            case 0: sort<0>(tid, getSource(), getSink());
                    break;
            case 1: sort<1>(tid, getSource(), getSink());
                    break;
            case 2: sort<2>(tid, getSource(), getSink());
                    break;
            case 3: sort<3>(tid, getSource(), getSink());
                    break;
         }
   }
};

