#include "fsbday.h"
#include "pipe.h"

class Store : public FSBdayTask<PipelineTask<Ale*> >
{
    public:
        void store(int level)
        {
            Ale* ale;

            uint32_t nparts = 0;

            AleSystem<ale_t, part_t> *alesys[2];

            alesys[0] = NULL;
            alesys[1] = NULL;

            if (this->getArgv()->is_final_level(level) && this->getArgv()->persist)
            {
                INFO_DEBUG(VERBOSITY > 0,
                    "Storing persistently! level: %u  list0: %u\n", level, this->getArgv()->get_list0(level, rank));
            }
            else
            {
                INFO_DEBUG(VERBOSITY > 0,
                    "Not storing persistently! level: %u  list0: %u\n", level, this->getArgv()->get_list0(level, rank));
            }


            if (getArgv()->is_final_level(level)) // We need only one alesys on the final level...
            {
                alesys[0] = new AleSystem<ale_t, part_t>(NPARTS, aledev, 
                    this->getArgv()->is_final_level(level) && this->getArgv()->persist, 
                    level, this->getArgv()->get_list0(0, rank) >> (level - this->getArgv()->startlevel));

                alesys[1] = alesys[0]; // I don't care if the list is even or odd...

                nparts = NPARTS;
            }
            else
            {
                alesys[0] = new AleSystem<ale_t, part_t>(NPARTS, aledev, 
                        this->getArgv()->is_final_level(level) && this->getArgv()->persist, 
                        level, this->getArgv()->get_list0(level, rank));

                alesys[1] = new AleSystem<ale_t, part_t>(NPARTS, aledev, 
                        this->getArgv()->is_final_level(level) && this->getArgv()->persist, 
                        level, this->getArgv()->get_list0(level, rank) + 1);

                nparts = 2*NPARTS;
            }

            for(uint32_t part=0; part < nparts; )
            {
                ale = getSource()->get();

                if (ale->is_last)
                    part++;

                if (ale->nelems != 0)
                    alesys[ale->list]->write_ale(ale->part, ale->nelems, ale->data);

                ale->is_last = false;

                getSink()->put(ale);
            }

            delete alesys[0];

            if (!getArgv()->is_final_level(level)) // There is only one alesystem on the final level...
                delete alesys[1];

            INFO_DEBUG(VERBOSITY > 1, "store done level %i\n", level);
        }

        void init()
        {
            Argv *argv = getArgv();

            if (argv->startlevel == 0)
                store(0);
        }

        void do_job(int level, int tid)
        {
            store(level+1);
        }
};


